/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: action.c,v 1.14 2004/07/13 10:22:52 lwa Exp $";

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "lmtp.h"


void free_action(struct fstate *state) {
  struct actionq *act = state->actions;
  while(act) {
    struct actionq *next = act->next;
    free(act->arg);
    free(act);
    act = next;
  }
  state->actions = NULL;
}

int push_action(struct fstate *state, char *a, 
		int(*cmd)(struct fstate*, struct actionq*)) {
  struct actionq *act = malloc(sizeof(struct actionq));
  char *arg = strdup(a);
  if (act==NULL || arg == NULL) {
    free(act);
    free(arg);
    free_action(state);
    MemoryError(state);
    return FILTER_ERROR;
  }
  act->arg = arg;
  act->level = state->hlevel++;
  act->logmask = state->action_log;
  act->cmd = cmd;
  act->logit = 0;
  act->action[0] = 0;

  act->next = state->actions;
  state->actions = act;

  return FILTER_CONTINUE;
}

static int action_log(struct fstate *state, struct actionq *act) {
  logit(state, act, ACTION_ERROR, "%s", act->arg);
  return FILTER_ERROR;
}

int push_action_log(struct fstate *state, char *fmt, ...) {
  char a[ERRMSGMAX+1];
  va_list ap;

  va_start(ap, fmt);
  vsnprintf(a, ERRMSGMAX, fmt, ap);
  va_end(ap);
  a[ERRMSGMAX]=0;

  return push_action(state, a, action_log);
}


int flush_action(struct fstate *state) {
  struct actionq *act = state->actions;
  int nactions = -1;
  char err[sizeof(state->error)];
  int logit = 0;
  if (exiting(state)) /* got some signal ? */
    return 0;
  while(act) {
    int ret;
    nactions ++;
    state->hlevel = act->level;
    if (nactions == 1)
      memcpy(err, state->error, sizeof(err));

    if (act->cmd)
      ret = act->cmd(state, act);
    else
      ret = FILTER_CONTINUE;
    if (act->logit)
      logit = 1;
    if (nactions == 0 && ret == FILTER_ERROR)
      return -1;
    act = act->next;
  }
  if (nactions > 0) {
    memcpy(state->error, err, sizeof(err));
  }
  if (logit)
    logflush(state, state->actions);
  return 0;
}

static char *bounce_message(struct fstate *state, char *a) {
  unsigned char *message  = (unsigned char *)a;
  unsigned char *c;
  while (*message == '-') {
    while (*message && *message != ' ' && *message != '\t') message++;
    while (*message == ' ' || *message == '\t') message++;
  }
  c = message;
  while(*c) {
    if (*c & 0x80)  *c &= 0x7F; /* strip 8th bit */
    if (iscntrl(*c)) {
      *c = 0;
      break;
    } 
    c++;
  }
  return (char *)message;
}

int action_deliver(struct fstate *state, struct actionq *act) {
  if (sendmd(state, act->arg)==NULL) {
    char *vacation_msg = "";
    if (state->vacation) {
      vacation_msg = send_vacation(state, state->vacation);
      free(state->vacation);
      state->vacation = NULL;
    }
    if (act->arg[0])
      logit(state, act, ACTION_DELIVER,
	    "delivered into '%s'%s", act->arg, vacation_msg);
    else
      logit(state, act, ACTION_DELIVER, "delivered%s", vacation_msg);
    seterr(state, 0, NULL, NULL); /* clear error status */
    return FILTER_STOP;
  } else {
    return FILTER_ERROR;
  }
}

/* same as action_deliver without vacation */
int action_deliver0(struct fstate *state, struct actionq *act) {
  char *vacation = state->vacation;
  int ret;
  state->vacation = NULL;
  ret = action_deliver(state, act);
  state->vacation = vacation;
  return ret;
}

int action_drop(struct fstate *state, struct actionq *act) {
  if (act->arg[0])
    logit(state, act, ACTION_DROP, "dropped (%s)", act->arg);
  else
    logit(state, act, ACTION_DROP, "dropped");
  return FILTER_STOP;
}

int action_bounce(struct fstate *state, struct actionq *act) {
  char *message  = bounce_message(state, act->arg);
  if (!*message) {
    message=state->user_bounce_message;
    if (!message || !*message)
      message=state->bounce_message;
  }
  seterr(state, 550, "5.7.0",  "<%s>: %s", state->qrecipient, message);
  logit(state, act, ACTION_BOUNCE, "bounced (%s)", 
	act->arg[0] ? act->arg : message);
  state->logged_message = 1;
  return FILTER_STOP;
}

int action_moved(struct fstate *state, struct actionq *act) {
  char message[ERRMSGMAX];
  if (act->arg[0]==0) {
    strcpy(message, "user has moved, no forwarding address");
    seterr(state, 551, "5.1.6", "<%s>: %s", state->qrecipient, message);
  } else {
    if (strchr(act->arg, ' ')) {
      strncpy(message, act->arg, sizeof(message));
    } else {
      char *at;
      if (act->arg[0]== '@' &&
	  (at = strrchr(state->qrecipient, '@')) != NULL) {
	snprintf(message, sizeof(message), 
		 "user has moved, please try <%.*s%s>",
		 at - state->qrecipient, state->qrecipient, act->arg);
      } else {
	snprintf(message, sizeof(message),
		 "user has moved, please try <%s>", act->arg);
      }
    }
    message[sizeof(message) -1 ] = 0;
    seterr(state, 551, "5.1.1", "<%s>: %s", state->qrecipient, message);
  }
  logit(state, act, ACTION_BOUNCE, "bounced (%s)", message);
  state->logged_message = 1;
  return FILTER_STOP;
}

int action_trash(struct fstate *state, struct actionq *act) {
  if (sendmd(state, FOLDER_TRASH)==NULL) {
    char *message = bounce_message(state, act->arg);
    if (*message) {
      seterr(state, 550, "5.7.0",  "<%s>: %s", state->qrecipient, message);
      logit(state, act, ACTION_DELIVER|ACTION_BOUNCE,
	    "trashed (%s)", act->arg);
      state->logged_message = 1;
    } else {
      if (act->arg[0])
	logit(state, act, ACTION_DELIVER, "trashed (%s)", act->arg);
      else
	logit(state, act, ACTION_DELIVER, "trashed");
    }
    return FILTER_STOP;
  } else {
    return FILTER_ERROR;
  }
}

int action_forward(struct fstate *state, struct actionq *act) {
  if (act->arg[0]==0) {
    if (sendmd(state, "")==NULL)
      logit(state, act, ACTION_ERROR|ACTION_DELIVER,
	    "delivered (forward without destination)");
    return FILTER_ERROR;
  } else {
    if (sendmail(state, state->sender, act->arg)==NULL) {
      logit(state, act, ACTION_FORWARD, "forwarded to %s", act->arg);
      return FILTER_STOP;
    } else {
      logit(state, act, ACTION_ERROR|ACTION_BOUNCE, 
	    "forwarded to %s (error=%s)", act->arg, state->error);
      return FILTER_STOP;
    }
  }
}

int action_forward2(struct fstate *state, struct actionq *act) {
  if (act->arg[0]==0) {
    logit(state, act, ACTION_ERROR, "ignored (forward2 without destination)");
    return FILTER_ERROR;
  } else {
    if (state->forwards_max >= 0 && state->forwards >= state->forwards_max) {
      logit(state, act, ACTION_ERROR, "ignored (too many forward2)");
      return FILTER_ERROR;
    }
    state->forwards ++;
    if (sendmail(state, state->sender, act->arg)==NULL) {
      logit(state, act, ACTION_FORWARD, "copy to %s", act->arg);
      return FILTER_CONTINUE;
    } else
      logit(state, act, ACTION_ERROR, 
	    "copy to %s (error=%s)", act->arg, state->error);
      return FILTER_ERROR;
  }
}

#define ARGV_NMAX 1000

int action_pipe(struct fstate *state, struct actionq *act) {
  char *path = acl_mailbox(&(state->pipe_path), state->rcpt->mbox);

  if (act->arg[0]==0) {
    if (sendmd(state, "")==NULL)
      logit(state, act, ACTION_ERROR|ACTION_DELIVER,
            "delivered (pipe without argument)");
    return FILTER_ERROR;
  } else {
    char *err = NULL;
    char args[ARG_MAX];
    char *argv[ARGV_NMAX];
    char *folder = NULL;
    int first_arg = 0;


    if (split_args(act->arg, args, sizeof(args), argv, ARGV_NMAX) < 0) 
      err = "no pipe command";

    if (!err && argv[0][0] == '-')
      folder=argv[first_arg++]+1;
    
    if (!err && (path == NULL || *path == 0))
      err = "pipe not allowed";

    if (!err && (argv[first_arg] == NULL || *argv[first_arg] == 0))
      err = "no pipe command";

    if (err && folder == NULL)
      folder = ""; /* Mail with syntax or perm error in INBOX */

    if (!err && 
	(err = send_pipe(state, folder, &argv[first_arg], path)) == NULL) {
      if (folder) {
	char *vacation_msg = "";
	if (state->vacation) {
	  vacation_msg = send_vacation(state, state->vacation);
	  free(state->vacation);
	  state->vacation = NULL;
	}
	if (*folder) {
	  logit(state, act, ACTION_DELIVER,
		"delivered into '%s' (piped via %s)%s", 
		folder, act->arg, vacation_msg);
	} else {
	  logit(state, act, ACTION_DELIVER,
		"delivered (piped via %s)%s", act->arg, vacation_msg);
	}
      } else {
	logit(state, act, ACTION_DELIVER, "piped to %s", act->arg);
      }
      return FILTER_STOP;
    } else {
      if (folder) {
	if (sendmd(state, folder) == NULL) {
	  if (*folder) {
	    logit(state, act, ACTION_ERROR|ACTION_DELIVER,
		  "delivered into '%s' (pipe '%s' has error: %s)",
		  folder, act->arg, err);
	  } else {
	    logit(state, act, ACTION_ERROR|ACTION_DELIVER,
		  "delivered (pipe '%s' has error: %s)", act->arg, err);
	  }
	} else {
	  logit(state, act, ACTION_ERROR|ACTION_DROP,
		"dropped (pipe '%s' has error: %s)", act->arg, err);
	}
      }
      return FILTER_STOP;
    }
  }
}

int action_av_notify(struct fstate *state, struct actionq *act) {
  char *path = act->arg;
  if (*path == '/') {
    if (av_message(state, path) == -1) { /* unable to notify */
      char *box = state->default_dest ? state->default_dest : "" ;
      if (strchr(box, '@'))
	return cmd_forward(state, box);
      else
	return cmd_deliver(state, box);
    }
  }
  logit(state, act, ACTION_DROP,
	"dropped (infected by %s)", state->av.virus);
  return FILTER_STOP;
}

int action_unknown(struct fstate *state, struct actionq *act) {
  if (sendmd(state, "")==NULL)
    logit(state, act, ACTION_ERROR|ACTION_DELIVER,
	  "delivered (bad command '%s')", act->arg);
  return FILTER_STOP;
}

