/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


static char const rcsid[] =
"$Id: av-drweb.c,v 1.13 2004/01/30 15:02:33 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/uio.h>

/* needed to avoid warning in FreeBSD */
#include <netinet/in.h>
/* POSIX path for htonl & Co */
#include <arpa/inet.h>

#include <unistd.h>
#include <syslog.h>

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "lmtp.h"


#define SIGNATURE "scanned by Doctor Web, http://www.drweb.ru/"

#define  DRWEBD_SCAN_CMD     1 /* scan file or buffer */
#define  DRWEBD_VERSION_CMD  2 /* return DrWeb Daemon version */
#define  DRWEBD_BASEINFO_CMD 3 /* return information on antivirus bases
                                  loaded by DrWeb Daemon */
#define  DRWEBD_IDSTRING_CMD 4 /* return information on DrWeb Daemon */
#define  DRWEBD_SCANPART_CMD 5 /* scan file fragment */


#define DRWEBD_RETURN_VIRUSES   0x0001 /* return names of the detected
					  viruses */
#define DRWEBD_RETURN_REPORT    0x0002 /* return scan log */
#define DRWEBD_RETURN_CODES     0x0004 /* return extended codes */
#define DRWEBD_HEURISTIC_ON     0x0008 /* use heuristic analyzer */
#define DRWEBD_CURE_FILES       0x0010 /* cure file */
#define DRWEBD_SPAM_FILTER      0x0020 /* use header filter */
#define DRWEBD_INFECTED_CURE    0x0040 /* try curing infected file; if fails
                                          the file is treated as incurable */
#define DRWEBD_INFECTED_MOVE    0x0080 /* move infected files */
#define DRWEBD_INFECTED_RENAME  0x0100 /* rename infected files */
#define DRWEBD_INFECTED_DELETE  0x0200 /* delete infected files */
#define DRWEBD_INCURABLE_MOVE   0x0400 /* move incurable files */
#define DRWEBD_INCURABLE_RENAME 0x0800 /* rename incurable files */
#define DRWEBD_INCURABLE_DELETE 0x1000 /* delete incurable files */
#define DRWEBD_SUSPECTED_MOVE   0x2000 /* move suspicious files */
#define DRWEBD_SUSPECTED_RENAME 0x4000 /* rename suspicious files */
#define DRWEBD_SUSPECTED_DELETE 0x8000 /* delete suspicious files */


#define DERR_NOERROR            0x00000000 /* no errors */
#define DERR_READ_ERR           0x00000001 /* read error(s), may occur both
                                              because of file physical
                                              inaccessibility (e.g., wrong
                                              access rights) and of logical
                                              damage of the document */
#define DERR_WRITE_ERR          0x00000002 /* write error(s) (possible only
                                              when curing) */
#define DERR_NOMEMORY           0x00000004 /* no memory */
#define DERR_CRC_ERROR          0x00000008 /* integrity of the scanned
                                              archive(s) has been broken */
#define DERR_READSOCKET         0x00000010 /* error in the log */
#define DERR_KNOWN_VIRUS        0x00000020 /* virus(es) found */
#define DERR_UNKNOWN_VIRUS      0x00000040 /* virus(es) suspected (possible
                                              only if heuristic analyzer is
                                              enabled) */
#define DERR_VIRUS_MODIFICATION 0x00000080 /* virus modification(s) found */
#define DERR_FILE_IS_CURED      0x00000100 /* file(s) cured */
#define DERR_TIMEOUT            0x00000200 /* scan timeout has run out */
#define DERR_SYMLINK            0x00000400 /* file(s) skipped because
					      symlink */
#define DERR_NO_REGFILE         0x00000800 /* file(s) skipped because not
                                              regular file */
#define DERR_SKIPPED            0x00001000 /* file(s) skipped, for example
                                              in ByType mode */
#define DERR_TOO_BIG            0x00002000 /* file(s) skipped because
                                              MaxFileSizeToExtract exceeded */
#define DERR_TOO_COMPRESSED     0x00004000 /* file(s) skipped because
                                              MaxCompressionRatio exceeded */
#define DERR_BAD_CALL           0x00008000 /* wrong command */
#define DERR_EVAL_KEY           0x00010000 /* DrWeb Daemon operates with demo
                                              key */
#define DERR_SPAM               0x00020000 /* file(s) denied by the header
                                              filter (possible only if header
                                              filter is enabled) */



static int read32(int ipc, uint32_t *lp) {
  int ret;
  uint32_t nl;
  ret = read(ipc, &nl, 4);
  *lp = ntohl(nl);
  return ret;
}

static int av_drweb_scan(struct fstate *state) {
  uint32_t status;
  int i;
  int fd ;
  char outbuf[AV_DESCR_MAX];
  char *outbuf_end;
  char *outbuf_c;
  static struct {
    uint32_t cmd;
    uint32_t option;
    uint32_t buffer;
    uint32_t maillen;
  } question;

  fd = rxconnect(&state->av.drweb);
  if (fd == -1)
    return -1;
  question.cmd     = htonl(DRWEBD_SCAN_CMD);
  question.option  = htonl(DRWEBD_RETURN_VIRUSES);
  question.buffer  = htonl(0);
  question.maillen = htonl(mailwrite_size(state, state->hdr_my_received,
					  HEADER_OTHER));

  if (write(fd, &question, sizeof(question)) == -1 ||
      mailwrite(state, state->hdr_my_received, HEADER_OTHER, fd) == -1 ||
      read32(fd, &status) == -1
      ) {
    roll_server(&state->av.drweb, "IO error in DrWeb server %s: %s",
		current_server(&state->av.drweb), strerror(errno));
    return -1;
  }

  if (status & (DERR_READ_ERR|DERR_WRITE_ERR|DERR_NOMEMORY|
		DERR_CRC_ERROR|DERR_TIMEOUT|DERR_BAD_CALL)) {
    roll_server(&state->av.drweb, "DrWeb server %s returns %08x",
		current_server(&state->av.drweb), status);
    return -1;
  }

  outbuf_c = outbuf;
  outbuf_end = outbuf + sizeof(outbuf) - 1;
  outbuf_c += snprintf(outbuf_c, outbuf_end - outbuf_c, "%s", SIGNATURE);

  if (read32(fd, &i) == -1) {
    roll_server(&state->av.drweb, "IO error in DrWeb server %s: %s", 
		current_server(&state->av.drweb), strerror(errno));
    return -1;
  }
  state->av.nvirus = i;

  if (i > 0) {
    char readbuf[AV_DESCR_MAX * 2];
    char *readbuf_end;
    char *readbuf_c;
    int len = 0;

    char infected[AV_DESCR_MAX];
    char *infected_c = infected;
    char *infected_end = infected + AV_DESCR_MAX;

    while(len < sizeof(readbuf)) {
      int readlen = read(fd, readbuf + len, sizeof(readbuf) - len);
      if (readlen == 0)
	break;
      if (readlen == -1) /* read error */
	break;
      len += readlen;
    }

    readbuf_c = readbuf;
    readbuf_end = readbuf + len;

    while(readbuf_c + 4 < readbuf_end && outbuf_c < outbuf_end && i-->0) {
      uint32_t namelen;
      int sl;
      namelen = ntohl(*(uint32_t *)readbuf_c);
      if (namelen == 0 || readbuf_c + namelen > readbuf_end )
	break;
      readbuf_c += 4;
      sl = snprintf(outbuf_c, outbuf_end - outbuf_c,
		    "\n\t%.*s", (int)namelen, readbuf_c);
      if (sl < 0) 
	break;
      outbuf_c += sl;
#define VANNOUNCE "infected with "
      if ((readbuf_c + sizeof(VANNOUNCE) -1 < readbuf_end) &&
	  infected_c < infected_end &&
	  strncmp(readbuf_c, VANNOUNCE, sizeof(VANNOUNCE) -1) == 0) {
	const char *format = ((infected_c == infected) ? "%.*s" : ", %.*s");
	/* warning with gcc -Wall on HPUX, snprintf prototype is bugus and
	   take a "char *" instead of "const char *" as third argument */
	sl = snprintf(infected_c, infected_end - infected_c, format,
		      (int)(namelen - sizeof(VANNOUNCE) - 1),
		      readbuf_c + sizeof(VANNOUNCE) - 1);
	if (sl > 0)
	  infected_c += sl;
      }
      readbuf_c += namelen;
    }

    if (infected_c > infected) {
      *infected_end = 0;
      av_set_virus(state, "DrWeb", infected);
    }

    *outbuf_end = 0;
    if (outbuf_c > outbuf_end) {
      outbuf_c = outbuf_end-1;
      while(*outbuf_c == '\n' && outbuf_c > outbuf) {
	*outbuf_c-- = 0;
      }
    }
  }

  state->av.desc = strdup(outbuf);
  rxclose(&state->av.drweb);
  return state->av.nvirus;
}

av_scanner_t av_drweb_init(struct fstate *state) {
  rxinit(&state->av.drweb, 0, NULL);
  return av_drweb_scan;
}
