/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: cmd.c,v 1.36 2004/04/20 13:18:13 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <syslog.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "lmtp.h"


int cmd_deliver(struct fstate *state, char *a) {
  push_action(state, a, action_deliver);
  return FILTER_STOP;
}

int cmd_deliver2(struct fstate *state, char *a) {
  return push_action(state, a, action_deliver0);
}

int cmd_archive2(struct fstate *state, char *a) {
  char year[5];
  char month[3];
  char day[3];
  char mbox[128]; /* TODO: use a configurable limit */
  struct tm *xtm=usertime(state);
  sprintf(year, "%04d", xtm->tm_year + 1900);
  sprintf(month, "%02d", xtm->tm_mon + 1);
  sprintf(day, "%02d", xtm->tm_mday);
  strfmt(mbox, sizeof(mbox), a, "$y$m$d", year, month, day);
  return push_action(state, mbox, action_deliver0);
}

int cmd_archive(struct fstate *state, char *a) {
  cmd_archive2(state, a);
  return FILTER_STOP;
}

int cmd_drop(struct fstate *state, char *a) {
  push_action(state, a, action_drop);
  return FILTER_STOP;
}

int cmd_bounce(struct fstate *state, char *a) {
  push_action(state, a, action_bounce);
  return FILTER_STOP;
}

int cmd_trash(struct fstate *state, char *a) {
  push_action(state, a, action_trash);
  return FILTER_STOP;
}

int cmd_forward(struct fstate *state, char *a) {
  push_action(state, a, action_forward);
  return FILTER_STOP;
}

int cmd_forward2(struct fstate *state, char *a) {
  return push_action(state, a, action_forward2);
}

int cmd_moved(struct fstate *state, char *a) {
  push_action(state, a, action_moved);
  return FILTER_STOP;
}

int cmd_pipe(struct fstate *state, char *a) {
  push_action(state, a, action_pipe);
  return FILTER_STOP;
}

int cmd_score(struct fstate *state, char *a) {
  int s;
  switch(*a) {
  case '/':
    s = readscore(state, &a+1);
    if (s)
      state->score/=s;
    else
      state->score=0;
    break;
  case '*':
    state->score*=readscore(state, &a+1);
    break;
  case '%':
    s = readscore(state, &a+1);
    if (s)
      state->score%=s;
    else
      state->score=s;
    break;
  case '+':
  case '-':
    state->score+=readscore(state, &a);
    break;
  case '=':
    a++;
  default:
    state->score=readscore(state, &a);
    break;
  }
  fdebug(state, "  score: " SCORE_FORMAT "\n", state->score);
  return FILTER_CONTINUE;
}

int cmd_skipto(struct fstate *state, char *a) {
  strncpy(state->tag, a, sizeof(state->tag)-1);
  return FILTER_CONTINUE;
}

int cmd_logrotate(struct fstate *state, char *a) {
  struct stat sb;
  if (state->log_msg && state->log_msg[0] &&
      state->logname && stat(state->logname, &sb)!=-1) {
    int secs = 3600 * strtol(a, NULL, 10);
    if (secs <= 0 || state->date-sb.st_atime > secs) {
      int fd;
      if ((fd = logrotate(state))>0) {
	close(fd);
	fdebug(state, "  force log rotation");
	syslog(LOG_INFO, "log rotation asked by <%s>", state->qrecipient);
      }
    }
  }
  return FILTER_CONTINUE;
}

int cmd_vacation(struct fstate *state, char *a) {
  int l=0;
  char *s;
  char *end=NULL;
  if (state->vacationcache==NULL)
    return FILTER_CONTINUE;
  if (state->vacation) {
    end=strchr(state->vacation, '\0');
    if (end==0)
      return FILTER_CONTINUE;  /* something wrong */
  } else {
    if (!*a)
      return FILTER_CONTINUE; /* must begin with something */
  }
  l=strlen(a);
  s=realloc(state->vacation, l+(end-state->vacation)+2);
  if (s==NULL) {
    MemoryError(state);
    return FILTER_STOP;
  }
  end=s+(end-state->vacation);
  strcpy(end, a);
  end[l++]='\n';
  end[l]=0;
  state->vacation=s;
  return FILTER_CONTINUE;
}

#define TRACE_SEP ", "

int cmd_trace(struct fstate *state, char *a) {
  int la;
  int lt;
  while(*a) {
    if (isspace((int)*a))
      a++;
    else if (*a == '-') {
      int on = 1;
      char *end;
      if (*++a == '-') {
	a++; on=0;
      }
      end = a;
      while(*end && !isspace((int)*end)) end++;
      if (end != a) {
	struct headers *hl = state->hdrlist;
	while(hl) {
	  if (hl->dd_len == end - a &&
	      strncasecmp(a, hl->header, end - a) == 0) {
	    if (on)
	      hl->trace |= TRACE_ON;
	    else 
	      hl->trace &= ~TRACE_ON;
	  }
	  hl = hl->next;
	}
      }
      a=end;
    } else
      break;
  }
  if (!*a)
    return FILTER_CONTINUE;
  la = strlen(a);
  if (state->trace) {
    lt=strlen(state->trace);
    state->trace = realloc(state->trace, lt + la + sizeof(TRACE_SEP));
    if (state->trace != NULL)
      strcpy(state->trace+lt, TRACE_SEP);
    lt+=sizeof(TRACE_SEP)-1;
  } else {
    lt = 0;
    state->trace = malloc(la+1);
  }
  if (state->trace == NULL) 
      syslog(LOG_ERR, "not enough space for trace");
  else 
    strcpy(state->trace+lt, a);
  return FILTER_CONTINUE;
}

int cmd_timezone(struct fstate *state, char *a) {
  if (*a==0 || *a==':' || *a == '/' || strchr(a, '.'))
    a=NULL;
  if (state->timezone_user) {
    free(state->timezone_user);
    state->timezone_user=NULL;
  }
  if (a) {
    state->timezone_user=strdup(a);
  }
  return FILTER_CONTINUE;
}


int cmd_add(struct fstate *state, char *a) {
  char content[MAX_USER_HEADER_LEN];
  struct headers *newh;
  if (!isheader(a))
    return FILTER_CONTINUE;
  purify_header(a, content, MAX_USER_HEADER_LEN);
  if ((newh = setheader(state->hlevel, "%s", content))==NULL)
    return FILTER_CONTINUE;
  newh->trace |= TRACE_ON;
  newh->next = state->hdr_delivered->next;
  state->hdr_delivered->next = newh;
  state->nbheaders++;
  return FILTER_CONTINUE;
}

int cmd_antivirus(struct fstate *state, char *a) {
  char *s = a;
  int always = 1;
  int nvirus;
  if (*s == '-')
    s++;
  while(*s) {
    if (*s == ':') {
      *s = 0;
      break;
    }
    if (*s<33 || *s>126) { /* RFC 2822, 3.6.8 */
      a = NULL;
      break;
    }
    s++;
  }
  if (a == NULL || *a == 0)
    a = state->av_default_header;

  if (*a == '-') {
    a++; always = 0;
  }

  if (*a == 0)
    a = "X-Antivirus";

  nvirus = av_scan(state);
  if (state->av.desc == NULL)
    return FILTER_CONTINUE;
  if (nvirus > 0 || (always && nvirus == 0)) {
    struct headers *newh;
    if ((newh = setheader(state->hlevel, "%s: %s", a, state->av.desc))==NULL)
      return FILTER_CONTINUE;
    if (nvirus > 0) {
      newh->trace |= TRACE_ON;
      fdebug(state, "  virus %s\n", state->av.virus);
    }
    newh->next = state->hdr_delivered->next;
    state->hdr_delivered->next = newh;
    state->hdr_delivered->next = newh;
    state->nbheaders++;
  }
  return FILTER_CONTINUE;
}

int cmd_spamassassin(struct fstate *state, char *a) {
  spam_assassin(state, a);
  fdebug(state, "  spamassassin score: %.2f\n", state->spamassassin_score);
  return FILTER_CONTINUE;
}

int cmd_blacklist(struct fstate *state, char *a) {
  char *fmt = a;
  while(*fmt && !isspace((int)*fmt))
    fmt++;
  while(isspace((int)*fmt))
      *fmt++=0;
  do {
    struct blresult *rr;
    char *rbl = a;
    char *test = NULL;
    while(*a) {
      if (*a==',') {
	*a++ = 0;
	break;
      } else if (*a == '=') { 
	*a++ = 0;
	test = a;
      } else {
	a++;
      }
    }
    rr=checkbl(state, rbl);
    if ( rr !=NULL && 
	 (test == NULL || *test == 0 || strcmp(test, rr->a) == 0)) {
      char *s=NULL;
      char buf[1024];
      struct headers *newh;
      state->user_bounce_message = rr->txt;
      if (fmt && *fmt) {
	s=strfmt(buf, sizeof(buf), fmt,
		 "$q$r$dp$a$t",
		 rr->query, rr->rbl, rr->domain, rr->proof,
		 rr->a, rr->txt);
      }
      fdebug(state, "  blacklist: %s in %s (%s)\n", rr->a, rr->rbl, rr->txt);
      if (s && isheader(s)) {
	char content[MAX_USER_HEADER_LEN];
	purify_header(s, content, MAX_USER_HEADER_LEN);
	if ((newh = setheader(state->hlevel, "%s", s))!=NULL) {
	  newh->trace |= TRACE_ON;
	  newh->next = state->hdr_delivered->next;
	  state->hdr_delivered->next = newh;
	  state->nbheaders++;
	}
      }
    }
  } while(*a);

  return FILTER_CONTINUE;
}

int cmd_extension(struct fstate *state, char *a) {
  if (*a) {
    char *ext = strdup(a);
    if (ext) {
      free(state->rcpt->ext);
      state->rcpt->ext = ext;
    }
  } else {
    free(state->rcpt->ext);
    state->rcpt->ext = NULL;
  }
  return FILTER_CONTINUE;
}

int cmd_stow(struct fstate *state, char *a) {
  char *ext = state->rcpt->ext;
  char folder[100];
  struct stat sb;
  if (!ext)
    return FILTER_CONTINUE; /* continue filtering */

  lower(ext); /* hmm.. */
  while(*a=='.') a++;
  *folder = '.';
  if (*a) {
    snprintf(folder, sizeof(folder), ".%s.%s", a, ext);
  } else {
    snprintf(folder, sizeof(folder), ".%s", ext);
  }
  folder[sizeof(folder)-1]=0;
  getbox(folder+1);
  if (folder[1] == 0 || lstat(folder, &sb) == -1 || !S_ISDIR(sb.st_mode)) {
    return FILTER_CONTINUE; /* continue filtering */
  }
  return cmd_deliver(state, folder+1);
}

int cmd_log(struct fstate *state, char *a) {
  readlogtype(state, a);
  return FILTER_CONTINUE;
}

int cmd_syslog(struct fstate *state, char *a) {
  if (state->syslog_host) {
    free(state->syslog_host);
    state->syslog_host = NULL;
  }
  if (*a && my_syslog_check(state, a)==0)
    state->syslog_host = strdup(a);
  return FILTER_CONTINUE;
}

int cmd_default(struct fstate *state, char *a) {
  if (state->default_dest)
    free(state->default_dest);
  state->default_dest = strdup(a);
  return FILTER_CONTINUE;
}

int cmd_debug(struct fstate *state, char *a) {
  if (state->filter_debug == NULL) {
    state->filter_debug = debug_init(state);
  }
  return FILTER_CONTINUE;
}

int cmd_header_set(struct fstate *state, char *a) {
  uint32_t set = get_set(*a);
  struct headers *hdr;
  if (set == 0)
    return FILTER_CONTINUE;
  for (hdr = state->hdrlist; hdr; hdr=hdr->next)
    hdr->inset &= ~set;
  while(!isblank((int)*a)) a++;
  while(*a) {
    char *hname = a;
    while(*a && *a != ':' && *a != ' ' && *a != '\t')  a++;
    if (a > hname)
      for (hdr = state->hdrlist; hdr; hdr=hdr->next) {
 	if (hdr->dd_len == a - hname  &&
	    strncasecmp(hdr->header, hname, hdr->dd_len) == 0)
	  hdr->inset |= set;
      }
    while(*a == ':' || *a == ' ' || *a == '\t') a++;
  }
  return FILTER_CONTINUE;
}

int cmd_stop(struct fstate *state, char *a) {
  return FILTER_ENDFILE;
}

int unknown_cmd(struct fstate *state, char *cmd, char *a) {
  push_action(state, cmd, action_unknown);
  return FILTER_STOP;
}

int default_cmd(struct fstate *state) {
  char *box = state->default_dest ? state->default_dest : "" ;
  if (state->rcpt != &state->av.viruskeeper) {
    char *av_action = acl_mailbox(&(state->av.action), state->rcpt->mbox);
    if (av_action && *av_action) {
      if (av_scan(state) > 0) {
	if (*av_action == '/') {
	  push_action(state, av_action, action_av_notify);
	  return FILTER_STOP;
	}
	if (*av_action == 'd') { /* drop */
	  push_action(state, "", action_av_notify);
	  return FILTER_STOP;
	}
      }
    }
  }
  if (strchr(box, '@'))
    return cmd_forward(state, box);
  else 
    return cmd_deliver(state, box);
}

