/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

#if WITH_DB

static char const rcsid[] =
"$Id: quotadb_db3.c,v 1.10 2003/11/10 15:52:25 lwa Exp $";

#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <db.h>

#include "lmtp.h"

/* change magic if you change the struct */
#define QUOTADB_DB3_MAGIC 0x0002

struct quotadb_db3_t {
  int    magic;
  time_t date;
  int    code;
  int    nbchecks;
};

static void quotadb_db3_err(char *user, char *context, int ret) {
  if (ret == DB_NOTFOUND || ret==0)
    return;
  if (user) {
    syslog(LOG_INFO, "error: <%s> %s: %s", user, context, db_strerror(ret));
  } else {
    syslog(LOG_ERR, "error: %s: %s", context, db_strerror(ret));
  }
}

static void quotadb_db3_keyuser(DBT *key, char data[LINELEN], char *user) {
    char *euser=data;
    
    while(*user) {
      *euser++=isupper(*user) ? tolower(*user) : *user;
      user++;
    }
    key->data = data;
    key->size = euser - data;
}

int quotadb_db3_open(struct fstate *state) {
  DB *dbp;
  DB_ENV *dbenv = NULL;
  int ret;
  char *envdir = state->quota_db_dbenv;
  char *quotafile = state->quota_db_dbfile;

  if (envdir == NULL)
    envdir = state->spool_user.path;
  
  if (*quotafile == 0)
    quotafile = NULL;

  if (state->quota_db_timeout<=0) {
    state->quota_db = NULL;
    return -1;
  }

  if (quotafile) {
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
      quotadb_db3_err(NULL, "db_env_create", ret);
      return -1;
    }
    if ((ret = dbenv->open(dbenv, envdir, 
			   DB_CREATE | DB_INIT_CDB |
			   DB_INIT_MPOOL,
			   0))!=0) {
      quotadb_db3_err(NULL, "dbenv->open", ret);
      dbenv->close(dbenv, 0);
      return -1;
    }
  }    

  if ((ret = db_create(&dbp, dbenv, 0)) != 0) {
    quotadb_db3_err(NULL, "db_create", ret);
    if (dbenv)
      dbenv->close(dbenv, 0);
    return -1;
  }

  if ( (ret = 
#if DB_VERSION_MAJOR == 3 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 0)  
	dbp->open(dbp, quotafile, NULL, DB_HASH, DB_CREATE, 0600)
#else
	dbp->open(dbp, NULL, quotafile, NULL, DB_HASH, DB_CREATE, 0600)
#endif
	) != 0) {
    quotadb_db3_err(NULL, "db->open", ret);
    dbp->close(dbp, 0);
    if (dbenv)
      dbenv->close(dbenv, 0);
    return -1;
  }

  state->quota_db=dbp;
  state->quota_dbenv=dbenv;
  return 0;
}

void quotadb_db3_close(struct fstate *state) {
  if (state->quota_db) {
    state->quota_db->close(state->quota_db, 0);
    state->quota_dbenv->close(state->quota_dbenv, 0);
    state->quota_db=NULL;
    state->quota_dbenv=NULL;
  }
}

void quotadb_db3_del(struct fstate *state, char *user) {
  DB *db = state->quota_db;
  if (db) {
    int ret;
    DBT key={};
    char xuser[LINELEN];
    quotadb_db3_keyuser(&key, xuser, user);

    if ((ret=db->del(db, NULL, &key, 0))!=0)
      quotadb_db3_err(user, "db->del", ret);
  }
}

char *quotadb_db3_check(struct fstate *state, char *user) {
  DB *db = state->quota_db;
  if (db) {
    DBT key={};
    DBT value={};
    int ret;
    struct quotadb_db3_t qt;
    char xuser[LINELEN];

    quotadb_db3_keyuser(&key, xuser, user);
    value.data=&qt;
    value.flags=DB_DBT_USERMEM;
    value.ulen=sizeof(qt);

    if ((ret=db->get(db, NULL, &key, &value, 0))==0) {
      if (value.ulen != value.size || qt.magic != QUOTADB_DB3_MAGIC) {
	quotadb_db3_err(user, "db->get corruption", EINVAL);
	if ((ret=db->del(db, NULL, &key, 0)!=0))
	  quotadb_db3_err(user, "db->del", ret);
	return NULL;
      }
      if (state->date - qt.date < state->quota_db_timeout) {
        qt.nbchecks++;
        if (state->quota_maxchecks > 0 && qt.nbchecks > state->quota_maxchecks) {
          qt.code = 552;
        }
        if ((ret=db->put(db, NULL, &key, &value, 0))!=0) 
          quotadb_db3_err(user, "db->put", ret);
	return seterr(state, qt.code, "4.2.2",
		      "<%s>: Mailbox full -- Quota exceeded", user);
      } else {
	if ((ret=db->del(db, NULL, &key, 0))!=0)
	  quotadb_db3_err(user, "db->del", ret);
      }
    } else {
      quotadb_db3_err(user, "db->get", ret);
    }
  }
  return NULL;
}



void quotadb_db3_add(struct fstate *state, char *user, int code) {
  DB *db = state->quota_db;

  if (db) {
    DBT key={};
    DBT value={};
    int ret;
    struct quotadb_db3_t qt;

    char xuser[LINELEN];

    quotadb_db3_keyuser(&key, xuser, user);

    qt.magic = QUOTADB_DB3_MAGIC;
    qt.date  = state->date;
    qt.code  = code;
    qt.nbchecks = 1;

    value.data  = &qt;
    value.size  = sizeof(qt);

    if ((ret=db->put(db, NULL, &key, &value, 0))!=0)
      quotadb_db3_err(user, "db->put", ret);
  }
}

#endif
