/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"
#if WITH_LIBSIEVE

static char const rcsid[] =
"$Id: sieve.c,v 1.1 2003/01/29 15:11:49 lwa Exp $";

#include <sieve_interface.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "lmtp.h"


static int filtersieve_error(struct fstate *state, char *format, ...) {
  va_list ap;
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n', stderr);
  return FILTER_ERROR;
}


static int s_redirect(void *ac, void *ic, void *sc, void *mc,
		      const char **err) {
  sieve_redirect_context_t *rc = (sieve_redirect_context_t *) ac;
  struct fstate *state = mc;
  cmd_forward(state, rc->addr);
  return SIEVE_OK;
}
static int s_discard(void *ac, void *ic, void *sc, void *mc,
		     const char **err) {
  struct fstate *state = mc;
  cmd_drop(state, "");
  return SIEVE_OK;
}

static int s_reject(void *ac, void *ic, void *sc, void *mc,
		    const char **err) {
  sieve_reject_context_t *rc = (sieve_reject_context_t *) ac;
  struct fstate *state = mc;
  cmd_bounce(state, rc->msg);
  return SIEVE_OK;
}
static int s_fileinto(void *ac, void *ic, void *sc, void *mc,
		      const char **err) {
  sieve_fileinto_context_t *fc = (sieve_fileinto_context_t *) ac;
  struct fstate *state = mc;
  cmd_deliver(state, fc->mailbox);
  return SIEVE_OK;
}

static int s_keep(void *ac, void *ic, void *sc, void *mc,
		  const char **err) {
  sieve_keep_context_t *kc = (sieve_keep_context_t *) ac;
  struct fstate *state = mc;
  cmd_deliver(state, "");
  return SIEVE_OK;
}

static int s_size(void *mc, int *size) {
  struct fstate *state = mc;
  *size = mailwrite_size(state, state->hdr_my_received, HEADER_OTHER);
  return SIEVE_OK;
}

static int s_header(void *v, const char *phead, const char ***body) {
  struct fstate *state = v;
  struct headers *hl = state->hdrlist;
  static const char **headbody = NULL;
  static int allocated = -1;
  int phead_len = strlen(phead);
  const char **headbody_c = headbody;
  if (allocated != state->nbheaders + 1) {
    allocated = state->nbheaders + 1;
    headbody = realloc(headbody_c, allocated * sizeof(char*));
    if (headbody == NULL) {
      if (headbody_c)
	free(headbody_c);
      allocated = -1;
      return SIEVE_FAIL;
    }
    headbody_c = headbody;
  }

  while(hl) {
    if (hl->dd_len == phead_len &&
	strncasecmp(phead, hl->header, phead_len) == 0)
      *headbody_c++=hl->header + hl->dd_len + 1;
    hl = hl->next;
  }
  *headbody_c = NULL;
  *body = headbody;
  return SIEVE_OK;
}

static int s_envelope(void *v, const char *head, const char ***body) {
  struct fstate *state = v;
  static const char *ebody[2] = { NULL, NULL };
  if (strcasecmp(head, "from") == 0) {
    ebody[0] = state->sender;
  } else if (strcasecmp(head, "to") == 0) {
    ebody[0] = state->qrecipient;
  } else {
    ebody[0] = NULL;
  }
  return SIEVE_OK;
}

static int s_parse_error(int lineno, const char *msg, void *i, void *s) {
  fprintf(stderr, "line %d: %s\n", lineno, msg);
  return SIEVE_OK;
}
static int s_execute_error(const char *msg, void *i, void *s, void *m) {
  struct fstate *state = m;
  fprintf(stderr, "execute_error: %s\n", msg);
  return SIEVE_OK;
}

#define REGISTER_SIEVE_CALLBACK(f) \
  if ((res=sieve_register_##f(i, &s_##f) != SIEVE_OK)) {		\
    fclose(ffilter);							\
    sieve_interp_free(&i);						\
    return								\
      filtersieve_error(state, "sieve_register_"#f"() returns %d", res); \
  }


int filtersieve(struct fstate *state, char *file) {
  FILE *ffilter;
  int final = FILTER_CONTINUE;
  if (file!= NULL && (ffilter=fopen(file, "r"))!=NULL) {
    sieve_interp_t *i;
    sieve_script_t *s;
    int res;
    if ((res=sieve_interp_alloc(&i, state)) != SIEVE_OK)
      return filtersieve_error(state, "sieve_interp_alloc() returns %d", res);

    REGISTER_SIEVE_CALLBACK(redirect);
    REGISTER_SIEVE_CALLBACK(discard);
    REGISTER_SIEVE_CALLBACK(reject);
    REGISTER_SIEVE_CALLBACK(fileinto);
    REGISTER_SIEVE_CALLBACK(keep);
    REGISTER_SIEVE_CALLBACK(size);
    REGISTER_SIEVE_CALLBACK(header);
    REGISTER_SIEVE_CALLBACK(envelope);
    /*    REGISTER_SIEVE_CALLBACK(vacation); */
    /*    REGISTER_SIEVE_CALLBACK(mark); */
    /*    REGISTER_SIEVE_CALLBACK(notify); */
    REGISTER_SIEVE_CALLBACK(parse_error);
    REGISTER_SIEVE_CALLBACK(execute_error);
    if ((res = sieve_script_parse(i, ffilter, NULL, &s)) != SIEVE_OK) {
      final = filtersieve_error(state, "sieve_script_parse() returns %d", res);
    } else {
      if ((res = sieve_execute_script(s, state) != SIEVE_OK))
	final = filtersieve_error(state, 
				  "sieve_execute_script() returns %d", res);
      else
	final = FILTER_STOP;
      sieve_script_free(&s);
    }
    sieve_interp_free(&i);
    fclose(ffilter);
  }
  return final;
}

#endif
