/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: store.c,v 1.56 2004/06/25 12:29:02 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/mman.h>

#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>


#define BUFLEN 1024


#include "lmtp.h"

#define THEADER(x, y) {#x, sizeof(#x)-1, y, }

struct tagged_header {
  char *header;
  size_t len;
  int type;
} static tagged_header_list[] = {
  THEADER(Content-Type,  HEADER_CONTENT_TYPE),
  THEADER(Delivered-To,  HEADER_DELIVERED_TO),
  THEADER(Message-ID,    HEADER_MESSAGE_ID),
  THEADER(Subject,       HEADER_SUBJECT),
  THEADER(Received,      HEADER_RECEIVED),
  THEADER(Return-Path,   HEADER_RETURN_PATH),
  { NULL, 0, HEADER_OTHER }
};

static int header_type(char *h, int len) {
  struct tagged_header *t = tagged_header_list;
  while(t->header) {
    if (len == t->len && strncasecmp(t->header, h, len) == 0)
      return t->type;
    t++;
  }
  return HEADER_OTHER;
}

static char *stdhdr (struct fstate *state) {
  struct headers *h;
  char bufdate[100];
  char *quoted;

  if (!state->hdr_in_received) {
    state->hdr_in_received = state->hdrlist;
  }
  
  if (!state->messageid) {
    if ((h=setheader(HEADER_MESSAGE_ID, "Message-ID: <%lX-" MID_FMT  "@%s>",
		     (long)state->date,
		     state->counter, 
		     state->myhostname))==NULL)
      return MemoryError(state);
    state->messageid=h->header+h->dd_len+2;
    h->trace = TRACE_DEFAULT;
    h->next  = state->hdrlist;
    state->hdrlist=h;
    state->nbheaders++;
  }

  syslog(LOG_INFO, ID_FMT "message-id=%s", state->counter, state->messageid);

  strftime(bufdate, sizeof(bufdate)-1, STRFTIME_FMT,
	   localtime(&(state->date)));

  if ((h=setheader(HEADER_RECEIVED, 
		   "Received: from %s (%s)\n"
		   "\tby %s (%s) with LMTP id %lu.%03lu;\n"
		   "\t%s", 
		   state->lhlo ? state->lhlo : "unknown", state->herhostname,
		   state->myhostname, state->myname,
		   (unsigned long)state->pid, state->counter,
		   bufdate))==NULL)
    return MemoryError(state);
  h->next = state->hdrlist;
  state->hdr_my_received=state->hdrlist=h;

  if ((h=setheader(HEADER_DELIVERED_TO, "Delivered-To: "))==NULL)
    return MemoryError(state);
  h->next = state->hdrlist;
  state->hdr_delivered = state->hdrlist = h;
  
  if ((quoted=quote_addr821(state->sender))==NULL ||
      (h=setheader(HEADER_RETURN_PATH, "Return-Path: <%s>", quoted))==NULL) {
    free(quoted);
    return MemoryError(state);
  }

  free(quoted);
  h->next = state->hdrlist;
  state->hdrlist=h;

  state->nbheaders+=3;  /* on ajoute 3 entetes dans cette fonction */

  return NULL;
}

char *prepare(struct fstate *state) {
  char spoolfile[PATH_MAX+1];
  int fd;

  snprintf(spoolfile, sizeof(spoolfile), "%s/lmtpd.XXXXXX",
	   state->spool_user.path);

  if ((fd=mkstemp(spoolfile))==-1) {
    return DeliverySysError(state, "opening spool file");
  }
  unlink(spoolfile);

  state->spool=fdopen(fd, "r+");
  if (state->spool==NULL) {
    char *error=DeliverySysError(state, "opening spool file");
    close(fd);
    return error;
  }
  
  return NULL;
}

static char *store_flush (struct fstate *state, char *error) {
  char *map;
  if (error)
    return error;
  if (fflush(state->spool)!=0)
    error=DeliverySysError(state, "flushing spool file");
  if (error)
    return error;
  state->bodysize= ftell(state->spool);
  state->msgsize+= state->bodysize;

  syslog(LOG_INFO, ID_FMT "from=<%s>, size=%lu, nrcpt=%d",
	 state->counter, state->sender,
	 (unsigned long)state->msgsize, (int)state->nrcpts);

  rewind(state->spool);
  map = mmap(0, state->bodysize, PROT_READ,
	     MAP_SHARED, fileno(state->spool), 0);
  if (map == MAP_FAILED)
    return DeliverySysError(state, "mmap spool file");
  state->body = map;

  fclose(state->spool);
  state->spool = NULL;
  return NULL;
}

char *store (struct fstate *state) {
  int newline=1;
  char *error=NULL;
  int len;
  char *header = NULL;
  int nb_Received=0;
  struct headers **newhdr;

  newhdr=&state->hdrlist;

  /*
   * lit les enttes
   */

  while(1) {
    int dd_len;
    int trace = 0;
    int type;

    timeout(state, state->timeout_dataline, 421);
    header=readheader(stdin, &len, &dd_len);
    timeout(state, state->timeout_dataline, 0);
    if (exiting(state)) {
      error=state->error;
      if (header) free(header);
      break;
    }
    if (header==NULL) {
      error=MemoryError(state);
      break;
    }
    if (dd_len==0)
      break;

    type = header_type(header, dd_len);

    if (type == HEADER_RETURN_PATH) {
      free(header);
      header = NULL;
      continue;
    }

    *newhdr=(struct headers*)malloc(sizeof(struct headers));
    if (*newhdr == NULL) {
      free(header);
      return MemoryError(state);
    }

    if (type == HEADER_RECEIVED) {
      if (++nb_Received > state->max_Received) {
	error=seterr(state, 554, "5.4.6", "Too many hops");
	free(*newhdr);
	free(header);
	*newhdr = NULL;
	break;
      }
      if (state->hdr_in_received == NULL 
	  && (state->received_front == NULL 
	      || (regexec(state->received_front, header+dd_len+1,
			  0, NULL, 0) == 0))) {
	state->hdr_in_received = *newhdr;
      }
    } else  if (!state->messageid && type == HEADER_MESSAGE_ID) {
      state->messageid=header+dd_len+1;
      trace = TRACE_DEFAULT;
      while(isspace((int)state->messageid[0]))
	state->messageid++;
    } else if (type == HEADER_SUBJECT)
      trace = TRACE_DEFAULT;

    (*newhdr)->header = header;
    (*newhdr)->utf8   = NULL;
    (*newhdr)->dd_len = dd_len;
    (*newhdr)->len    = len;
    (*newhdr)->trace  = trace;
    (*newhdr)->next   = NULL;
    (*newhdr)->type   = type;
    (*newhdr)->inset  = 0;
    newhdr=&(*newhdr)->next;
    state->msgsize+=len+1;
    state->nbheaders++;
  }

  if (error==NULL && (error=stdhdr(state))==NULL && header) {
    state->nblines=state->nbheaders+1;
    if (*header && *header != '\n') {
      putc('\n', state->spool);
      state->nblines++;
    }
    fputs(header, state->spool);
    state->msgsize+=len;
    if (*header=='\0') { /* no body nor separation line */
      putc('\n', state->spool);  /* workaround for a microsoft reader bug */
      free(header);
      return store_flush(state, error);
    }
    free(header);
  }

  /*
   * lit le corp du message
   */ 

  while(1) {
    int nextnewline;
    char buf[BUFLEN], *ebuf;

    timeout(state, state->timeout_dataline, 421);
    if (fgets(buf, BUFLEN-1, stdin)==NULL) {
      timeout(state, state->timeout_dataline, 0);
      return seterr(state, 421, "4.5.0", "End of input flow");
    }
    timeout(state, state->timeout_dataline, 0);
    if (exiting(state)) {
      error = state->error;
      break;
    }

    ebuf=memchr(buf, '\n', BUFLEN-1);

    if (ebuf) {
      nextnewline=1;
      state->nblines++;
    } else {
      ebuf=buf+BUFLEN-1;
      nextnewline=0;
    }

    if (ebuf>buf && ebuf[-1]=='\r') {
      ebuf[-1]=ebuf[0];
      *ebuf--=0;
    }


    if (newline && *buf=='.') {
      if (buf[1]=='\n')
	return store_flush(state, error);
      if (!error && (len=fwrite(buf+1, 1, ebuf-buf, state->spool))==EOF)
	error=DeliverySysError(state, "writing spool file");
    }
    else
      if (!error && (len=fwrite(buf, 1, ebuf-buf+1, state->spool))==EOF)
	error=DeliverySysError(state, "writing spool file");
    
    newline=nextnewline;
  }

  return error; 
}
