/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: user.c,v 1.20 2003/10/16 11:23:24 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "lmtp.h"

#ifdef HAVE_LIBHOME
#include <hpwd.h>
#else
#include <pwd.h>
#endif

static struct rcpt *getsharedrcpt(struct fstate *state, char *domain,
				  struct rcpt *rcpt1) {
  FILE *f;
  char buf[1024];
  char xuser[1024];
  char *euser=domain;

  strncpy(xuser, rcpt1->user, euser - rcpt1->user);
  xuser[euser - rcpt1->user]=0;
  if ((euser=strchr(xuser, '+')) !=NULL )
    *euser = 0;
  if (!xuser[0])
    return NULL;
  if ((f=fopen(state->share_map, "r"))==NULL)
    return NULL;
  while(fgets(buf, sizeof(buf), f)) {
    char *q;
    if ((q=strchr(buf, '\n')) != NULL)
      *q=0;
    if ((q=strchr(buf, '#')) != NULL)
      *q=0;
    q=buf;
    while(*q) {
      if (isspace((int)*q))
	break;
      q++;
    }
    if (!q)
      break;
    *q++=0;
    if (strcasecmp(buf, xuser)==0) {
      struct stat sb;
      char mbox[1024];
      while(*q && isspace((int)*q))
	q++;
      if (!q)
	break;
      if (stat(q, &sb)==-1)
	break;
      if (!S_ISDIR(sb.st_mode))
	break;
      if (!sb.st_uid)
	break;
      snprintf(mbox, sizeof(mbox), "%s@%s", buf, state->share_domain);
      rcpt1->mbox   = strdup(mbox);
      rcpt1->path   = strdup(q);
      rcpt1->rcpt   = rcpt1->user;
      rcpt1->ext    = NULL;
      rcpt1->uid    = sb.st_uid;
      rcpt1->gid    = sb.st_gid;
      rcpt1->shared = 1;
      rcpt1->quota  = 0;
      fclose(f);
      return rcpt1;
    }
  }
  fclose(f);
  return NULL;
}

/*
 *  add localdomain to 's' when no domain
 */
static char *add_default_domain(struct fstate *state, char *s) {
  if (s && state->localdomain && strchr(s, '@') == NULL) {
    char *res;
    res = realloc(s, strlen(s) + 2 + strlen(state->localdomain));
    if (res) {
      s = strcat(res, "@");
      s = strcat(s, state->localdomain);
    }
  }
  return s;
}

static struct rcpt *getpwdrcpt(struct fstate *state, struct rcpt *rcpt1) {
  
  struct passwd *pwd;
  static int inited = 0;
  char *recipient, *extension;
  errno=0;

  if (delimit_extension(rcpt1->user, state->recipient_delimiter,
			&recipient, &extension) == -1) {
    MemoryError(state);
    return NULL;
  }

  if (quotadb_check(state, recipient)) {
    free(recipient);
    free(extension);
    return NULL;
  }

  if (!inited) {
#if HAVE_SETPASSENT
    setpassent(1);
#endif
    inited = 1;
  }

  if (state->localdomain) { /* remove localdomain if any */
    char *d = strchr(recipient, '@');
    if (d && strcasecmp(d+1, state->localdomain) == 0)
      *d = 0;
  }
  pwd=getpwnam(recipient);


  if (pwd==NULL) {
    if (extension) free(extension);
    if (errno == ENOMEM) {
      seterr(state, 451, "4.1.0", "<%s> system failure", rcpt1->user);
      return NULL;
    } else {
      seterr(state, 550, "5.1.1", "<%s> user unknown", rcpt1->user);
      return NULL;
    }
  }

  rcpt1->ext    = extension;
  rcpt1->mbox   = add_default_domain(state, strdup(pwd->pw_name));
  rcpt1->path   = strdup(pwd->pw_dir);
  rcpt1->rcpt   = add_default_domain(state, recipient);
  rcpt1->uid    = pwd->pw_uid;
  rcpt1->gid    = pwd->pw_gid;
  rcpt1->shared = 0;
#if HAVE_STRUCT_PASSWD_PW_QUOTA
  rcpt1->quota = pwd->pw_quota;
#else
  rcpt1->quota = 0;
#endif
  return rcpt1;
}

void freercpt1(struct rcpt *rcpt1) {
  if (rcpt1->user)
    free(rcpt1->user);
  rcpt1->user = NULL;
  if (rcpt1->path)
    free(rcpt1->path);
  rcpt1->path = NULL;
  if (rcpt1->mbox)
    free(rcpt1->mbox);
  rcpt1->mbox=NULL;
  if (rcpt1->ext)
    free(rcpt1->ext);
  rcpt1->ext=NULL;
  if (rcpt1->rcpt)
    free(rcpt1->rcpt);
  rcpt1->rcpt = NULL;
  if (rcpt1->bounce)
    free(rcpt1->bounce);
  rcpt1->bounce = NULL;
}

struct rcpt *getrcpt(struct fstate *state, struct rcpt *rcpt1) {
  
  if (state->share_domain && state->share_domain[0]) {
    char *domain = strchr(rcpt1->user, '@');
    if (domain && strcasecmp(state->share_domain, domain+1)==0) {
      if (getsharedrcpt(state, domain, rcpt1)) {
	return rcpt1;
      } else {
	seterr(state, 550, "5.1.1", "<%s> user unknown", rcpt1->user);
	return NULL;
      }
    }
  }

  if (getpwdrcpt(state, rcpt1))
    return rcpt1;

  return NULL;
}


char *newrcpt(struct fstate *state, char *user) {
  struct rcpt *rcpt1;

  rcpt1=(struct rcpt*)realloc(state->rcpts,
			      (state->nrcpts+1)*sizeof(struct rcpt));
  if (rcpt1==NULL) {
    return MemoryError(state);
  }

  state->rcpts=rcpt1;
  rcpt1+=state->nrcpts;
  rcpt1->user = user;
  rcpt1->bounce = NULL;
  rcpt1->path = NULL;
  rcpt1->mbox = NULL;
  rcpt1->rcpt = NULL;

  if (getrcpt(state, rcpt1)) {
    state->nrcpts++;
    return NULL;
  } else if (state->can_bounce && state->error[0] == '5') {
    if ((rcpt1->bounce = strdup(state->error)) != NULL) {
      state->nrcpts++;
      return NULL;
    }
  }

  return state->error;
}


void do_vrfy(struct fstate *state, char *user) {
  struct rcpt rcptx={};
  rcptx.user = user;

  quotadb_del(state, user);

  if (!getrcpt(state, &rcptx)) {
    fputs(state->error, stdout);
    return;
  }
  if (strcasecmp(rcptx.mbox, rcptx.user)==0) {
    printf("250 2.0.0 <%s>\r\n", rcptx.mbox);
  } else {
    printf("251 2.1.0 will forward to <%s>\r\n", rcptx.mbox);
  }
  freercpt1(&rcptx);
}

void do_expn(struct fstate *state, char *user) {
  struct rcpt rcptx={};
  rcptx.user = user;

  quotadb_del(state, user);

  if (!getrcpt(state, &rcptx)) {
    fputs(state->error, stdout);
    return;
  }
  printf("250 2.0.0 <%s>\r\n", rcptx.mbox);
  freercpt1(&rcptx);
}
