/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: xconnect.c,v 1.11 2004/01/30 14:58:20 lwa Exp $";

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <syslog.h>

/* struct timeval in non Posix include */
#if !HAVE_STRUCT_TIMEVAL_TV_SEC
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#endif

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "lmtp.h"

static int xxconnect(int s, struct sockaddr *sa, int len, int timeout) {

  struct timeval stimeout;
  int saved_errno;;

  if (timeout) {
    fcntl(s, F_SETFL, O_NONBLOCK);
    stimeout.tv_sec = timeout;
    stimeout.tv_usec = 0;
  }  

  if (connect(s, sa, len) == 0) {
    if (timeout)
      fcntl(s, F_SETFL, 0);
    return s;
  }

  saved_errno = errno;

  if (timeout == 0) {
    close(s);
    return -1;
  }

  fcntl(s, F_SETFL, 0);

  if (saved_errno == EINPROGRESS) {
    fd_set fdset;
    FD_ZERO(&fdset);
    FD_SET(s, &fdset);
    switch(select(s+1, NULL,  &fdset, NULL, &stimeout)) {
    case 0:
      saved_errno = ETIMEDOUT;
    case -1:
      break;
    default:
      return s;
    }
  }
  close(s);
  errno = saved_errno;
  return -1;
}

/* connect to the path socket */
/* interface may change to include timeout */

int xconnect(char *path, int timeout) {
  int domain;
  int s;
  char *service = path; /* dummy init for gcc -Wall */
  
  if (*path == 0) {
    errno = EINVAL;
    return -1;
  }  

  if (*path == '/') {
    domain = AF_UNIX;
  } else if ((service = strchr(path, ':')) != NULL) {
    domain = AF_INET;
  } else {
    errno = EPROTONOSUPPORT;
    return -1;
  }
  s = socket(domain, SOCK_STREAM, 0);
  if (s == -1)
    return -1;
  switch (domain) {
  case AF_UNIX: {
    struct sockaddr_un sa_un;
    memset(&sa_un, 0, sizeof(struct sockaddr_un));
    /*    sa_un.sa_un_len = 0; */
    sa_un.sun_family = AF_UNIX;
    sa_un.sun_path[0] = 0;
    strncat(sa_un.sun_path, path, sizeof(sa_un.sun_path)-1);
    return xxconnect(s, (struct sockaddr *)&sa_un, sizeof(sa_un), timeout);
  }
  case AF_INET: {
    struct sockaddr_in sa_in;
    char addr[1024];
    memset(&sa_in, 0, sizeof(struct sockaddr_in));
    addr[0] = 0;
    strncat(addr, path, service - path);
    service ++;
#if HAVE_INET_PTON
    inet_pton(AF_INET, addr, &sa_in.sin_addr);
#else
    sa_in.sin_addr.s_addr = inet_addr(addr);
#endif
    sa_in.sin_family = AF_INET;
    sa_in.sin_port = htons(strtoul(service, NULL, 10));
    return xxconnect(s, (struct sockaddr *)&sa_in, sizeof(sa_in), timeout);
  }
  default:
    close(s);
    return -1;
  }
  return s;
}

#define MAX_RANDOM_PERCENT (0x80000000UL/100)

char *current_server(struct rservers *servers) {
  return servers->arg.argv[servers->current];
}

static void rxclose1(struct rservers *servers) {
  if (servers->fd != -1) {
    if (servers->close)
      servers->close(servers->fd);
    close(servers->fd);
    servers->fd = -1;
  }
}

int roll_server(struct rservers *servers, char *reason, ...) {
  rxclose1(servers);
  servers->current ++;
  if (servers->current >= servers->arg.argc)
    servers->current = 0;
  if (reason) {
    va_list ap;
    va_start(ap, reason);
    vsyslog(LOG_WARNING, reason, ap);
    va_end(ap);
  }
  return servers->current;
}

int rxconnect(struct rservers *servers) {
  int init;
  if (servers->arg.argc <= 0) {
    errno = ERANGE;
    return -1;
  }
  
  init = servers->current;

  while(servers->fd == -1) {
    char *s = current_server(servers);
    servers->fd = xconnect(s, servers->timeout);
    if (servers->fd == -1) {
      if (roll_server(servers, "unable to connect to server '%s': %s",
		      s, strerror(errno))
	  == init) 
	return -1;
    }
  }
  return servers->fd;
}

int rxconnect_test(struct rservers *servers, char *buf, int len) {
  if (servers->fd != -1) {
    if (write(servers->fd, buf, len) == -1) {
      roll_server(servers, NULL);
    } else {
      return servers->fd;
    }
  }
  if (rxconnect(servers) == -1)
    return -1;

  if (write(servers->fd, buf, len) == -1) {
    roll_server(servers, "unable to write to server %s: %s",
		current_server(servers), strerror(errno));
    return -1;
  }

  return servers->fd;
}

int rxclose(struct rservers *servers) {
  if (!servers->keepalive)
    rxclose1(servers);

  if (servers->arg.argc > 1 &&
      servers->roll_percent > 0 &&
      (servers->roll_percent > 100 ||
       servers->roll_percent > (random() / MAX_RANDOM_PERCENT)))
    roll_server(servers, NULL);
  return 0;
}

void rxinit(struct rservers *servers, int keepalive, void (*close_f)(int)) {
  servers->fd = -1;
  servers->keepalive = keepalive;
  servers->close = close_f;
  servers->current = ((double)random() / 0x80000000UL) * servers->arg.argc;
}
